/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : SetCoverSolution.cc                                      */
/*                                                                       */
/*       Date : April     12, 1997                                       */
/* Updated at : September 12, 1998                                       */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#include <tSet.h>
#include <SortedSet.h>
#include <deffunc.h>

#include "Column.h"
#include "SetCoverProblem.h"
#include "SetCoverSolution.h"
#include "SetCover.h"

/*************************************************************************/
/** Standard SetCoverSolution object constructor.  */
/*************************************************************************/

SetCoverSolution::SetCoverSolution(SetCoverProblem *newProblem)
{  
   setProblem(newProblem);

}// end of constructor

/*************************************************************************/
/** A constructor for SetCoverSolution object for dynamic creation.  */
/*************************************************************************/

SetCoverSolution::SetCoverSolution()
{  myProblem = NULL;
   myCover = NULL;
   weights = NULL;
   cost    = 0;
}// end of constructor

/*************************************************************************/
/** A Copy constructor for SetCoverSolution Object. We need deep copy.  */
/*************************************************************************/

SetCoverSolution::SetCoverSolution(const SetCoverSolution& RHS)
{  
   copyFrom(RHS);

}// end of COPY constructor

/*************************************************************************/
/*************************************************************************/

void SetCoverSolution::copyFrom(const SetCoverSolution& RHS)
{
   //cout << " In SetCoverSolution::copyFrom() " << endl;
   Column *c;
   int i, j;
   int numofRows ;
   int numofCols ;

   myProblem = RHS.myProblem ;
   Solution = RHS.Solution;
   cost    = RHS.cost;

   if( myProblem!=NULL )
   {
      numofRows = myProblem->getNumOfRows();
      numofCols = myProblem->getNumOfCols();
      myCover = new Column(numofRows);
      weights = new int[numofRows];

      for(i=0;i<numofRows;i++) if( RHS.myCover->test(i) ) myCover->set(i);
      for(i=0;i<numofRows;i++) weights[i]=0;

      cost = 0;

      for(i=0;i<getNumOfColumns();i++)
      if( (c=getColumnWithIndx(i)) != NULL )
      {
          for(j=0;j<numofRows;j++) if( c->test(j) ) weights[j]++;
	  cost += c->getProfit();

      }// end of if

   }// end of if

   //cout << " Out of SetCoverSolution::copyFrom() " << endl;

}// end of COPY constructor

/*************************************************************************/
/**
   Destructor for the SetCoverSolution object. Here we need to 
   deallocate a pointer array.
*/
/*************************************************************************/

SetCoverSolution::~SetCoverSolution()
{  
   dispose();

}// end of destructor

/*************************************************************************/
/*************************************************************************/

void SetCoverSolution::dispose()
{
   //cout << " In SetCoverSolution::dispose() " << endl;
   if(myCover!=NULL) delete myCover;
   if(weights!=NULL) delete[] weights;
   myCover = NULL;
   weights = NULL;
   cost    = 0;
   //cout << " Out of SetCoverSolution::dispose() " << endl;

}// end of dispose()

/*************************************************************************/
/** sets the problem for this solution.  */
/*************************************************************************/

void SetCoverSolution::setProblem(SetCoverProblem *newProblem)
{  
   myProblem = newProblem;
   Solution.clearAll();
   myCover = NULL;
   weights = NULL;
   cost    = 0;

   if( myProblem!=NULL )
   {
      int i, j;
      int numofRows = myProblem->getNumOfRows();
      int numofCols = myProblem->getNumOfCols();
      myCover = new Column(numofRows);
      weights = new int[numofRows];
      for(i=0;i<numofRows;i++) weights[i]=0;

   }// end of if

}// end of setProblem()

/*************************************************************************/
/**
   Returns the pointer of the Set Cover problem that this solution
   belongs to.
*/
/*************************************************************************/

SetCoverProblem * SetCoverSolution::getProblem() const
{  
   return myProblem;

}// end of getProblem()

/*************************************************************************/
/** Returns the number of columns in this solution.  */
/*************************************************************************/

unsigned SetCoverSolution::getNumOfColumns() const
{  
   return Solution.getSize();

}// end of getNumOfColumns() 

/*************************************************************************/
/** Returns the number of columns in this solution.  */
/*************************************************************************/

unsigned SetCoverSolution::getNumberOfOverlaps() const
{  
   int ret = 0;

   for(int i=0;i<myProblem->getNumOfRows();i++) 
       if(weights[i]>1) ret += (weights[i]-1);

   return ret;

}// end of getNumberOfOverlaps() 

/*************************************************************************/
/** Returns the cover without excluding any column.  */
/*************************************************************************/
char *SetCoverSolution::getCover() const { return getCover(-1); }

/*************************************************************************/
/** Returns the cover by excluding column exCol.  */
/*************************************************************************/
char *SetCoverSolution::getCover(int exCol) const
{  
   int  NumOfRows = myProblem->getNumOfRows();
   char *ret = new char[NumOfRows];
   char *SolId;
   unsigned i, j;
   Column *clm;

   for(j=0;j<NumOfRows;j++) ret[j]=0;

   // for each column
   for(i=0;i<getNumOfColumns();i++)
   {
      clm = getColumnWithIndx(i);
      if( i!=exCol && clm!=NULL ) 
      {
	 SolId = clm->getCover();

	 if( SolId!=NULL )
	 {

           for(j=0;j<NumOfRows;j++) if(SolId[j]==1) ret[j]=1;

	   delete[] SolId;

	 }// end of if

      }// end of if

   }// end of for

   return ret;

}// end of getCover()

/*************************************************************************/
/** Calculates the currently covered rows without the given column.  */
/*************************************************************************/
char *SetCoverSolution::getCoverWithoutColIndex(int indx)
{
   return getCover(indx);

}// end of getCoverWithoutColIndex()

/*************************************************************************/
/** Returns the column object representing cover.  */
/*************************************************************************/
Column SetCoverSolution::getCoverColumn(int indx) const
{
   //cout << " In SetCoverSolution::getCoverColumn() " << endl;
   Column ret( myProblem->getNumOfRows());
   char *cover = getCover(indx);
   Column a = ret.getColumn(cover, myProblem->getNumOfRows() );
   delete[] cover;
   //cout << " Out of SetCoverSolution::getCoverColumn() " << endl;
   return a;

}// end of getCoverColumn()

/*************************************************************************/
/*************************************************************************/
tSet<int> 
SetCoverSolution::getColumns() { return Solution.getOnlyValues(); }
//SetCoverSolution::getColumns() { return Solution; }

/*************************************************************************/
/** Adds a new column to the solution */
/*************************************************************************/
void SetCoverSolution::addColumn(size_t columnIndex)
{  
   int i;

   // Update Cover, Cost, and Weights
   Column *c = myProblem->getColumnWithIndx(columnIndex);
   if( c!=NULL )
   {
      Solution.addElement(columnIndex,c->getProfit());
      //Solution.addElement(columnIndex);

      cost += c->getProfit();
      for(i=0;i<myProblem->getNumOfRows();i++)
	if( c->test(i) ) { myCover->set(i); weights[i]++; }

   }// end of if

}// end of addColumn()

/*************************************************************************/
/** deletes the indxTH column index from the solution set. */
/*************************************************************************/

void SetCoverSolution::deleteColumnWithIndex(size_t indx)
{  
   int i;

   // Update Cover, Cost, and Weights
   Column *c = getColumnWithIndx(indx);
   if( c!=NULL )
   {
      Solution.removeAt(indx);
      //Solution.removeElementAt(indx);

      cost -= c->getProfit();
      for(i=0;i<myProblem->getNumOfRows();i++)
	if( c->test(i) ) { weights[i]--; if(weights[i]<1) myCover->reset(i); }

   }// end of if

}// end of deleteColumnWithIndex()

/*************************************************************************/
/**
   returns the address of the corresponding column in the solution
   set.
*/
/*************************************************************************/

Column* SetCoverSolution::getColumnWithIndx(int indx) const
{  
   int cIndx ; 

   return (Solution.getAt(indx,cIndx)?
   //return (Solution.getElementAtSafe(indx,cIndx)?
	    myProblem->getColumnWithIndx(cIndx):NULL);

}// end of getColumnWithIndx()

/*************************************************************************/
/*************************************************************************/

SetCoverSolution::getColNumWithIndex(int indx) const
{
   int cIndx;
   return (Solution.getAt(indx,cIndx)?cIndx:-1);
   //return (Solution.getElementAtSafe(indx,cIndx)?cIndx:-1);

}// end of getColNumWithIndex()

/*************************************************************************/
/**
   returns the total cost of this solution. It is sum of the column
   costs.
*/
/*************************************************************************/

double SetCoverSolution::getCost() const {  return cost; }

/*************************************************************************/
/**
    returns the number of nonzeros in the solutions upto given
    column(inclusive).
*/
/*************************************************************************/

SetCoverSolution::getNumOfNonZerosUpTo(int indx)
//{  int maxi = maximum(indx,(*myProblem).getNumOfRows());
{  
   unsigned int ret =0;

   /*
   int maxi = maximum(indx,index);
   for(int i=0;i<maxi;i++)
      if(Solution[i]!=NULL) 
	   ret+=(unsigned int)(*Solution[i]).getNumOfNonZero();
   */

   return ret;
       
}// end of getNumOfNonZerosUpTo()

/*************************************************************************/
/** pushes the empty ColumnPair slots end of the Solution array */
/*************************************************************************/

void SetCoverSolution::organize()
{  unsigned int i, j;
   Column *dummy;
   ulong maxi = (*myProblem).getNumOfRows();

   /*
   i=0;
   while(i<maxi-1)
   {  
      // if ith element is NULL we need to put something here

      if(Solution[i]==NULL)
      {  
	 j=i+1;  // Start search from the next element on the list

	 // find first non-NULL element in the array

         while(j<maxi&&Solution[j]==NULL) j++;

	 if(j<maxi)                    // Solution[j]!=NULL
	 {  Solution[i] = Solution[j];
	    Solution[j] = NULL;
	 }// end of if 

      }// end of if

      i++;         // goto next element on the array

   }// end of while
   */

}// end of organize()

/*************************************************************************/
/**
   finds a solution randomly 
   starts from the given column index
*/
/*************************************************************************/

void SetCoverSolution::restore(int start)
{

   unsigned int i=0, j;
   ulong NumOfRows = myProblem->getNumOfRows();
   ulong NumOfCols = myProblem->getNumOfCols();

   Column current(NumOfRows);
   Column cvr = getCoverColumn();
   Column *cj;

   j = start;

   while(i<NumOfCols && cvr.count()<NumOfRows)
   {
       if( (cj=myProblem->getColumnWithIndx(j)) != NULL )
       {
          if( (*cj-cvr).count() > 0 )
          { 
	     addColumn(j);
             cvr |= *cj ;

          }// end of if

       }// end of if

       j = (j+1) % NumOfCols;

       i++;

   }// end of while

   reduce();

}// end of restore()

/*************************************************************************/
/** finds a solution randomly according to greedy strategy */
/*************************************************************************/

void SetCoverSolution::solveRandom(int size)
{  
   /*************************/
   /** Start from the fresh */
   /*************************/
   Solution.clearAll();
   restore( rand() % size);

}// end of solveRandom()

/*************************************************************************/
/*************************************************************************/

SetCoverSolution SetCoverSolution::operator=(const SetCoverSolution& RHS)
{  
   dispose();
   copyFrom(RHS);
   return (*this);

}// end of operator=()

/*************************************************************************/
/**
    We can not compare the order of bits. We should compare the columns
    in the solution and this produces the "SET" so that difference set
    gives us the difference information between solutions.
*/
/*************************************************************************/

SetCoverSolution::operator==(const SetCoverSolution& RHS)
{  
   int ret = ( myProblem == RHS.myProblem );

   ret = ret && ( Solution == RHS.Solution );

   return ret;

}// end of operator=()

/*************************************************************************/
/** Prints all the columns in binary format */
/*************************************************************************/

void SetCoverSolution::printColumnsBinary(ostream& os) const
{
   //cout << " in SetCoverSolution::printColumnsBinary() " << endl;
   int i;
   Column *c;
   int indx;

   os << myProblem->getNumOfRows() << " "
      << getNumOfColumns() << endl;

   for(i=0;i<getNumOfColumns();i++)
      if( (c = getColumnWithIndx(i))!=NULL) os << *c << endl;

}// end of printColumnsBinary()

/*************************************************************************/
/** Prints all the columns in binary format */
/*************************************************************************/

void SetCoverSolution::printColumns(ostream& os) const
{
   //cout << " in SetCoverSolution::printColumns() " << endl;
   int i;
   Column tmp(myProblem->getNumOfRows());
   Column cover = getCoverColumn();
   Column *c;

   os << " cover : "; cover.printBinary(os); os<< endl;

   int tot = 0;
   for(i=0;i<getNumOfColumns();i++)
      if((c = getColumnWithIndx(i))!=NULL) 
      { os << i << " : " << *c << endl; tot += c->count(); } 
 
   os << endl << " Overlaps : " << (tot-myProblem->getNumOfRows()) << endl;

}// end of printColumns()

/*************************************************************************/
/** Prints the solution */
/*************************************************************************/

ostream &operator<<(ostream& os, const SetCoverSolution& RHS)
{  
    //cout << " in &operator<<() " << endl;

    os << " # of Rows : " << (*(RHS.myProblem)).getNumOfRows();
    os << " # of Columns : " 
       << (*(RHS.myProblem)).getNumOfCols() << endl << endl;

    RHS.printColumns(os);

    os << " The solution's cost is : " << RHS.getCost() << endl << endl;
    os << RHS.Solution << endl << endl;

    os << " End of Print " << endl;

    return os ;

}// end of operator<<()

/*************************************************************************/
/*************************************************************************/

Column* SetCoverSolution::findMinimumCostRatio(Column* U) 
{
   float mincost = 100000.0;
   float mcost = 100000.0;
   Column *ret = NULL;
   int i,j,k,numofcovered,id;
   Column  *c;
   tSet<int> Res;

   //cout << " in SetCoverSolution::findMinimumCostRatio() " << endl;
   for(i=0;i<myProblem->getNumOfRows();i++)
   {
      if(U->test(i)) // if not covered
      {
	 //cout << " i: " << i 
	 //     << " beforemyProblem->getColumnSetCoveringRow(i) " << endl;
	 Res.clearAll();
         myProblem->getColumnSetCoveringRow(i,&Res);

	 //cout << " Res " << Res.getSize() << endl;

         for(j=0;j<Res.getSize();j++)
	 {
	    if( Res.getElementAtSafe(j,id) )
	    {
	       c = myProblem->getColumnWithIndx(id);

	       if( c!=NULL && c->test(i))
	       {
	          numofcovered=0;
	          for(k=0;k<myProblem->getNumOfRows();k++)	
		     if(U->test(k)&&c->test(k)) numofcovered++;

		  //cout << " j: " << j << "check " << id << endl;
	          mcost = c->getProfit()/numofcovered;

	          if(mcost<mincost) { mincost=mcost; ret=c; }

	       }// end of if

	    }// end of if

	 }// end of for
	  
      }// end of if

   }// end of for

   //cout << " out of SetCoverSolution::findMinimumCostRatio() " << endl;

   return ret;

}// end of findMinimumCostRatio

/*************************************************************************/
/*************************************************************************/

Column* SetCoverSolution::findMinimumCostRatioCHAVATAL(Column* U) 
{
   float mincost = 100000.0;
   float mcost = 100000.0;
   Column *ret = NULL;
   int i,j,k,numofcovered,id;
   Column  *c;
   tSet<int> Res;

   //cout << " in SetCoverSolution::findMinimumCostRatio() " << endl;
   for(i=0;i<myProblem->getNumOfRows();i++)
   {
      if(U->test(i)) // if not covered
      {
	 //cout << " i: " << i 
	 //     << " beforemyProblem->getColumnSetCoveringRow(i) " << endl;
	 Res.clearAll();
         myProblem->getColumnSetCoveringRow(i,&Res);

	 //cout << " Res " << Res.getSize() << endl;

         for(j=0;j<Res.getSize();j++)
	 {
	    if( Res.getElementAtSafe(j,id) )
	    {
	       c = myProblem->getColumnWithIndx(id);

	       if( c!=NULL && c->test(i) && c->getProfit()<mincost ) 
	       { mincost=c->getProfit(); ret=c; }


	    }// end of if

	 }// end of for
	  
      }// end of if

   }// end of for

   //cout << " out of SetCoverSolution::findMinimumCostRatio() " << endl;

   return ret;

}// end of findMinimumCostRatio

/*************************************************************************/
/*************************************************************************/

Column* SetCoverSolution::findMinimumCostRatioGRDY3(Column* U) 
{
   float mincost = 100000.0;
   float mcost = 100000.0;
   Column *ret = NULL;
   int i,j,k,numofcovered,id,numOverlap;
   Column  *c;
   tSet<int> Res;

   //cout << " in SetCoverSolution::findMinimumCostRatio() " << endl;
   for(i=0;i<myProblem->getNumOfRows();i++)
   {
      if(U->test(i)) // if not covered
      {
	 //cout << " i: " << i 
	 //     << " beforemyProblem->getColumnSetCoveringRow(i) " << endl;
	 Res.clearAll();
         myProblem->getColumnSetCoveringRow(i,&Res);

	 //cout << " Res " << Res.getSize() << endl;

         for(j=0;j<Res.getSize();j++)
	 {
	    if( Res.getElementAtSafe(j,id) )
	    {
	       c = myProblem->getColumnWithIndx(id);

	       if( c!=NULL && c->test(i))
	       {
	          numofcovered=0;
	          for(k=0;k<myProblem->getNumOfRows();k++)	
		     if(U->test(k)&&c->test(k)) numofcovered++;

	          numOverlap=0;
	          for(k=0;k<myProblem->getNumOfRows();k++)	
		     if(!U->test(k)&&c->test(k)) numOverlap++;

		  //cout << " j: " << j << "check " << id << endl;
	          mcost = c->getProfit()/numofcovered;
		  if(numOverlap>0) mcost *= numOverlap;

	          if(mcost<mincost) { mincost=mcost; ret=c; }

	       }// end of if

	    }// end of if

	 }// end of for
	  
      }// end of if

   }// end of for

   //cout << " out of SetCoverSolution::findMinimumCostRatio() " << endl;

   return ret;

}// end of findMinimumCostRatio

/*************************************************************************/
/**  Solves with Greedy Approach */
/*************************************************************************/

void SetCoverSolution::solveWithGreedy(GREEDY_TYPE gType) 
{
   int i;
   Column  U(myProblem->getNumOfRows());
   Column  *c;
   tSet<int> Res;
   double total  =0.0;

   /** None is covered */
   for(i=0;i<myProblem->getNumOfRows();i++) U.set(i);

   int stay=1;
   while( U.count() && stay )
   {
      if(gType==CHAVATAL_GRDY) 
	 c = findMinimumCostRatioCHAVATAL(&U);
      else if(gType==GRDY_3) 
	 c = findMinimumCostRatioGRDY3(&U);
      else 
	 c = findMinimumCostRatio(&U);

      if(c!=NULL)
      {
         //cout << " Candidate : " << *c << endl;
         Res.addElement(c->getIndex());
	 total += (double) c->getProfit();
	 //cout << " Res " << Res << endl;

         //for(i=0;i<myProblem->getNumOfRows();i++) if(c->test(i)) U.reset(i);
	 U -= *c;

	 //cout << " U " << U << endl;

      } else { cout << " c is null " << endl; stay=0; }

      //cout << " uncovered " << U << endl;

   }// end of while

   //cout << " Greedy solution is " << Res << endl ;
   //cout << " with " << total << " cost " << endl;

   int indx;

   Solution.clearAll();
   for(i=0;i<Res.getSize();i++)
      if(Res.getElementAtSafe(i,indx)) addColumn(indx); 

}// end of solveWithGreedy()

/*************************************************************************/
/**  Solves with Random Choice Approach */
/*************************************************************************/

void SetCoverSolution::solveWithRandomChoice(int start) 
{
   int i, a, j;
   Column  U(myProblem->getNumOfRows());
   Column  *c = NULL;
   tSet<int> Res;
   double total  =0.0;
   int    numOfRows = myProblem->getNumOfRows();

   // Clear the Solution
   Solution.clearAll();

   // None is covered 
   for(i=0;i<numOfRows;i++) U.reset(i);

   for(j=0;j<numOfRows;j++)
   {
    i = (start+j) % numOfRows;
    if( !U.test(i) )
    {
       Res.clearAll();
       //cout << " before getColumnSetCoveringRow() " << endl;
       myProblem->getColumnSetCoveringRow(i,&Res);
       if( Res.getSize()>0 )
       {
          //cout << " random pick  " << endl;
	  a = rand() % Res.getSize();

	  if( Res.getElementAtSafe(a,a) )
	  {   
	      c = myProblem->getColumnWithIndx(a); 
	      if( c!=NULL )
	      {
	         addColumn(a); 
		 //cout << " before U |= *c; " << endl;
                 //for(j=0;j<myProblem->getNumOfRows();j++) 
		 //    if( c->test(j) ) U.set(j);
	         U |= *c;
		 //cout << " after U |= *c; " << endl;
		 total += c->getProfit();
		 //cout << " after c->getProfit();" << endl;

	      }// end of if

	  }// end of if

       }// end of if

       //cout << " uncovered " << U << endl;

    }// end of if
   }// end of for

   //cout << " Random Choice solution is " << Res << endl ;
   //cout << " with " << total << " cost " << endl;

}// end of solveWithGreedy()

/*************************************************************************/
/** Removes unneccassary columns */
/*************************************************************************/
/*
void SetCoverSolution::reduce()
{
   int i, stay, j;
   Column *ci;
   Column *cj;
   int indx;

   //cout << " In SetCoverSolution::reduce() " << endl;
   int initial=i=getNumOfColumns()-1;
   while(i>=0)
   {
      //cout << " i: " << i << endl;
      if( (ci = getColumnWithIndx(i))!=NULL) 
      {
	 stay = 1;
         j=i-1;

         while(j>=0&&stay)
           if( (cj = getColumnWithIndx(j))!=NULL &&  (*ci & *cj) == *ci ) stay=0;
           else j--; 

	 // remove ci
	 if(!stay) deleteColumnWithIndex(i);
	 else i--;

      } else i--;

   }// end of for

   //cout << " reduced from " << initial+1 << " to " << getNumOfColumns() << endl;

}// end of reduce()
*/

/*************************************************************************/
/** Removes unneccassary columns */
/*************************************************************************/

void SetCoverSolution::reduce()
{
   int i, stay, j;
   Column *ci;
   Column *cj;
   int indx;
   int numOfRows = myProblem->getNumOfRows();
   int numOfCols = myProblem->getNumOfCols();
   int *w = new int[numOfCols];

   cout << " reduce() " << endl;
   for(i=0;i<numOfCols;i++) w[i]=0;

   for(i=0;i<getNumOfColumns();i++)
      if( (ci = getColumnWithIndx(i))!=NULL) 
	  for(j=0;j<numOfCols;j++) if( ci->test(j) ) w[j]++;


   /* If it is not SortedSet 
   Column** cls = new Column*[getNumOfColumns()];
   for(i=0;i<getNumOfColumns();i++) cls[i] = getColumnWithIndx(i);
   int *sindex = new int[getNumOfColumns()];

   for(i=0;i<getNumOfColumns();i++) sindex[i]=i;

   for(i=0;i<getNumOfColumns()-1;i++)
     for(j=i+1;j<getNumOfColumns();j++)
     {
	if(cls[sindex[i]]->getProfit()>cls[sindex[j]]->getProfit())
	{ stay = sindex[j];
	  sindex[j] = sindex[i];
	  sindex[i] = stay;
	}
     }

   tSet<int> aSolution(getNumOfColumns());

   for(i=0;i<getNumOfColumns();i++) 
       aSolution.addElement(getColNumWithIndex(sindex[i]));

   delete[] sindex;
   delete[] cls;

   Solution = aSolution;
   */

   i=getNumOfColumns()-1;

   while(i>-1)
   {
     // decide whether to remove it or not
     stay=1;

     if( (ci = getColumnWithIndx(i))!=NULL) 
     {
        stay=0;
	for(j=0;!stay&&j<numOfCols;j++) 
	    if( ci->test(j) && w[j]<2 ) stay=1;
	
     }// end of if

     if(!stay) // buckle up 
     {
	for(j=0;j<numOfCols;j++) if( ci->test(j) ) w[j]--;
        deleteColumnWithIndex(i);
	cout << " column " << i << " is deleted! " << endl;

     }// end of if

     i--;

   }// end of while

   //cout << " reduced from " << initial+1 << " to " << getNumOfColumns() << endl;
   delete[] w;

}// end of reduce()

/*************************************************************************/
/** */
/*************************************************************************/

SetCoverSolution::restoreRandomChoice(tSet<int>* nlegal)
{

   //cout << " in of SetCoverSolution::restoreRandomChoice(int size) " << endl;
   ulong NumOfRows = myProblem->getNumOfRows();
   int   ret = NumOfRows;
   unsigned int i=0, indx, j;
   Column cover = getCoverColumn();
   tSet<int> R1;
   Column *c;
   int id;

   for(i=0;i<NumOfRows;i++)
   {
      if( !cover.test(i) )
      {
	 R1.clearAll();
	 myProblem->getColumnSetCoveringRow(i,&R1);
	 if( nlegal!=NULL ) R1 -= *nlegal;
	 if( R1.getSize() )
	 {
	   indx = rand() % R1.getSize();

	   id = -1;
	   if( R1.getElementAtSafe(indx,id) && 
	       (c=myProblem->getColumnWithIndx(id))!=NULL )
	   {
	       addColumn(id);
	       cover |= *c;

	   }// end of if

	 }// end of if

      }// end of if

   }// end of for

   ret = NumOfRows - cover.count();

   if( ret==0 ) reduce();

   //cout << " out of SetCoverSolution::restoreRandomChoice(int size) " << endl;

   return ret;

}// end of restoreRandomChoice()

/*************************************************************************/
/** */
/*************************************************************************/

SetCoverSolution::restoreGreedyChoice(tSet<int>* nlegal)
{

   //cout << " in of SetCoverSolution::restoreGreedyChoice(int size) " << endl;
   ulong NumOfRows = myProblem->getNumOfRows();
   int   ret = NumOfRows;
   unsigned int i=0, indx, j;
   Column cover = getCoverColumn();
   tSet<int> R1;
   Column *c;
   int id;

   for(i=0;i<NumOfRows;i++)
   {
      if( !cover.test(i) )
      {
	 R1.clearAll();
	 myProblem->getColumnSetCoveringRow(i,&R1);
	 if( nlegal!=NULL ) R1 -= *nlegal;
	 if( R1.getSize() )
	 {
	   indx = 0;

	   id=-1;
	   if( R1.getElementAtSafe(indx,id) && 
	       (c=myProblem->getColumnWithIndx(id))!=NULL )
	   {
	       addColumn(id);
	       cover |= *c;

	   }// end of if

	 }// end of if

      }// end of if

   }// end of for

   ret = NumOfRows - cover.count();

   if( ret==0 ) reduce();

   //cout << " out of SetCoverSolution::restoreGreedyChoice(int size) " << endl;

   return ret;

}// end of restoreGreedyChoice()

/*************************************************************************/
/** 
    Returns -1 if Column is not found.

    CHAVATAL_GRDY : First
    RANDOM        : random
    BACK_GRDY     : per unit cost
    GRDY_3        : minimize the overlaps with per unit cost

*/
/*************************************************************************/

SetCoverSolution::pickChoice( Column * cover, tSet<int>* R1, GREEDY_TYPE gType)
{
   int ret = 0;
   float mcost, tcost;
   int i, id, numofoverlap, numofcovered, k ;
   Column *c;

   switch(gType)
   {
     case CHAVATAL_GRDY : ret = 0; 
                          id=-1;
			  mcost = FLT_MX;

	                  if( R1->getElementAtSafe(ret,id) && 
	                      (c=myProblem->getColumnWithIndx(id))!=NULL )
			     mcost = c->getProfit();

		          i=1;
                          while(i<R1->getSize())
		          {
                             id=-1;
	                     if( R1->getElementAtSafe(i,id) && 
	                         (c=myProblem->getColumnWithIndx(id))!=NULL )
			         tcost = c->getProfit();
			     
			     if(tcost<mcost) { mcost=tcost; ret=i; }
			     i++;

		          }// end of while

			  break;

     case RANDOM        : ret = rand() % R1->getSize(); break;

     case BACK_GRDY     : ret = 0;

		     mcost = FLT_MX;
                     id=-1;

	             if( R1->getElementAtSafe(ret,id) && 
	                 (c=myProblem->getColumnWithIndx(id))!=NULL )
                     {
	                numofcovered=0;
	                for(k=0;k<myProblem->getNumOfRows();k++)	
		           if(!cover->test(k)&&c->test(k)) numofcovered++;

			 if(numofcovered<1) numofcovered=1;
	                 mcost = c->getProfit()/numofcovered;

                     }// end of if

		     i=1;
                     while(i<R1->getSize())
		     {
                       id=-1;
	               if( R1->getElementAtSafe(i,id) && 
	                   (c=myProblem->getColumnWithIndx(id))!=NULL )
                       {
	                   numofcovered=0;
	                   for(k=0;k<myProblem->getNumOfRows();k++)	
		              if(!cover->test(k)&&c->test(k)) numofcovered++;

			   if(numofcovered<1) numofcovered=1;

	                   tcost = c->getProfit()/numofcovered;

			   if(tcost<mcost) { mcost=tcost; ret=i; }

                       }// end of if

		       i++;

		     }// end of while

		     break;

     case GRDY_3        : ret = 0;

		     mcost = FLT_MX;
                     id=-1;

	             if( R1->getElementAtSafe(ret,id) && 
	                 (c=myProblem->getColumnWithIndx(id))!=NULL )
                     {

			 numofcovered = (*c-*cover).count();
			 if(numofcovered<1) numofcovered=1;

	                 numofoverlap=0;
	                 for(k=0;k<myProblem->getNumOfRows();k++)	
		           if(cover->test(k)&&c->test(k)) numofoverlap++;

	                 tcost = c->getProfit()/numofcovered;
		         if(numofoverlap>0) tcost *= numofoverlap;

	                 mcost=tcost; 

                     }// end of if

		     i=1;
                     while(i<R1->getSize())
		     {

		       id=-1;
	               if( R1->getElementAtSafe(i,id) && 
	                   (c=myProblem->getColumnWithIndx(id))!=NULL )
                       {

			   numofcovered = (*c-*cover).count();
			   if(numofcovered<1) numofcovered=1;

	                   numofoverlap=0;
	                   for(k=0;k<myProblem->getNumOfRows();k++)	
		             if(cover->test(k)&&c->test(k)) numofoverlap++;

	                   tcost = c->getProfit()/numofcovered;
		           if(numofoverlap>0) tcost *= numofoverlap;

			   if(tcost<mcost) { mcost=tcost; ret=i; }

                       }// end of if

		       i++;

		     }// end of while

		     break;

   }// end of switch()

   return ret;

}// end of pickChoice()

/*************************************************************************/
/**
    Restores the solution according to given algorithm with abandon
    list.
*/
/*************************************************************************/

SetCoverSolution::restoreG(tSet<int>* nlegal, GREEDY_TYPE gType)
{

   //cout << " in of SetCoverSolution::restoreG(int size) " << endl;
   ulong NumOfRows = myProblem->getNumOfRows();
   int   ret = NumOfRows;

   unsigned int i=0, indx, j;
   Column cover = getCoverColumn();
   tSet<int> R1;
   Column *c;
   int id;

   for(i=0;i<NumOfRows;i++)
   {
      if( !cover.test(i) )
      {
	 R1.clearAll();

	 myProblem->getColumnSetCoveringRow(i,&R1);

	 if( nlegal!=NULL ) R1 -= *nlegal;

	 if( R1.getSize() )
	 {

	     // Pick a column from R1 w.r.t. algorithm
	   indx = pickChoice(&cover,&R1,gType);

	   id=-1;
	   if( R1.getElementAtSafe(indx,id) ) 
	   {
	     if( (c=myProblem->getColumnWithIndx(id))!=NULL )
	     {
	        addColumn(id);
	        cover |= *c;

	     }// end of if

	   }// end of if

	 }// end of if

      }// end of if

   }// end of for

      // Find number of uncovered rows
   ret = NumOfRows - cover.count();

      // Restore the Solution without abandon list.
   if( nlegal!=NULL && ret>0)    restoreG(NULL,gType);

      // Column reduction in the solution
   //if( ret==0 ) reduce();

   //cout << " out of SetCoverSolution::restoreG(int size) " << endl;

   return ret;

}// end of restoreG()

/*************************************************************************/
/*************************************************************************/

SetCoverSolution::restoreRandom(tSet<int>* nlegal)
{
   return restoreG(nlegal,RANDOM);

}// end of restoreRandom()

/*************************************************************************/
/*************************************************************************/

SetCoverSolution::restoreBack(tSet<int>* nlegal)
{
   return restoreG(nlegal,BACK_GRDY);

}// end of restoreBack()

/*************************************************************************/
/*************************************************************************/

SetCoverSolution::restoreChavatal(tSet<int>* nlegal)
{
   return restoreG(nlegal,CHAVATAL_GRDY);

}// end of restoreChavatal()

/*************************************************************************/
/*************************************************************************/

SetCoverSolution::restoreGreedy3(tSet<int>* nlegal)
{
   return restoreG(nlegal,GRDY_3);

}// end of restoreGreedy3()

